"use strict";(self.webpackChunkmitosheet=self.webpackChunkmitosheet||[]).push([[922],{922:(e,t,n)=>{n.r(t),n.d(t,{MODULE_NAME:()=>l,MODULE_VERSION:()=>s,default:()=>a});const i=n(147),s=i.version,l=i.name,a=n(624).default},147:e=>{e.exports=JSON.parse('{"files":["lib/**/*.js","fonts/**/*","dist/*.js","css/**/*.css","style/index.js"],"jupyterlab":{"outputDir":"mitosheet/labextension","extension":"lib/plugin"},"name":"mitosheet","license":"AGPL-3.0-only","author":{"name":"Mito Sheet","email":"naterush1997@gmail.com"},"bugs":{"url":"https://github.com/mito-ds/monorepo/issues"},"repository":{"url":"https://github.com/mito-ds/monorepo","type":"git"},"version":"0.1.578","dependencies":{"@jupyterlab/application":"^3.5.0","@jupyterlab/notebook":"^3.0.6","@types/fscreen":"^1.0.1","@types/react-dom":"^17.0.2","fscreen":"^1.1.0","react":"^17.0.1","react-dom":"^17.0.1","streamlit-component-lib":"^2.0.0"},"scripts":{"clean:lib":"rimraf lib tsconfig.tsbuildinfo","clean:nbextension":"rimraf mitosheet/nbextension","clean:labextension":"rimraf mitosheet/labextension","clean:mito_frontend":"rimraf mitosheet/mito_frontend.js mitosheet/mito_frontend.css","clean:dash":"cd mitosheet && rimraf mitodashwrapper.css && rimraf mitodashwrapper.js","clean:all":"jlpm run clean:lib && jlpm run clean:nbextension && jlpm run clean:labextension && jlpm run clean:mito_frontend && npm run clean:dash","clean":"npm run clean:all","build:nbextension:dev":"webpack --mode=development","build:labextension:dev":"jupyter labextension build --development True .","build:mito_frontend:dev":"node esbuild.js","build:streamlit:dev":"node esbuild.streamlit.js","build:all:dev":"npm run build:nbextension:dev && build:labextension:dev && build:mito_frontend:dev && build:streamlit:dev","build:dev":"npm run build:all:dev","build:lib":"tsc","build:nbextension":"webpack --mode=production","build:labextension":"jupyter labextension build .","build:mito_frontend":"node esbuild.js","build:streamlit":"node esbuild.streamlit.js","build:dash":"node esbuild.dash.js","build:all":"npm run build:lib && npm run build:labextension && npm run build:nbextension && npm run build:mito_frontend && npm run build:streamlit && npm run build:dash","build":"npm run build:all","watch:lib":"tsc -w","watch:nbextension":"webpack --watch --mode=development","watch:labextension":"jupyter labextension watch .","watch:mito_frontend":"node esbuild.js --watch","watch:streamlit":"node esbuild.streamlit.js --watch","watch:dash":"node esbuild.dash.js --watch","watch:all":"run-p watch:lib watch:labextension watch:nbextension watch:mito_frontend watch:streamlit watch:dash","watch":"npm run watch:all","lint:check":"eslint src/ --ext .ts,.tsx","lint":"eslint src/ --ext .ts,.tsx --fix","prepack":"npm run build:all","build:docs:update_frontend":"python docs/make_function_docs.py update_frontend","build:docs:generate_markdown":"python docs/make_function_docs.py generate_markdown"},"keywords":["jupyter","jupyterlab","jupyterlab-extension","widgets","ipython","ipywidgets"],"devDependencies":{"@jupyterlab/builder":"^3.0.0","@types/expect.js":"^0.3.29","@types/node":"^18.11.18","@types/react":"^17.0.26","@typescript-eslint/eslint-plugin":"^4.8.1","@typescript-eslint/parser":"^4.8.1","esbuild":"^0.15.16","eslint":"^7.14.0","eslint-config-prettier":"^8.3.0","eslint-plugin-office-addins":"^1.0.3","eslint-plugin-prettier":"^4.0.0","eslint-plugin-react":"^7.21.3","expect.js":"^0.3.1","fs-extra":"^10.0.0","mkdirp":"^1.0.4","mocha":"^9.1.2","npm-run-all":"^4.1.5","prettier":"^2.1.1","rimraf":"^3.0.2","typescript":"^4.4.3"},"main":"lib/index.js","homepage":"https://trymito.io","types":"./lib/index.d.ts","description":"The Mito Spreadsheet","resolutions":{"@types/react":"^17.0.2"}}')}}]);