define((()=>(()=>{"use strict";var e={632:(e,o,t)=>{var n,i;t.r(o),e=t.hmd(e),window.__webpack_public_path__=(null===(n=document.querySelector("body"))||void 0===n?void 0:n.getAttribute("data-base-url"))+"nbextensions/mitosheet",window.require&&window.require.config({map:{"*":{mitosheet:"nbextensions/mitosheet/index"}}}),null===(i=window.Jupyter)||void 0===i||i.toolbar.add_buttons_group([{id:"mito-toolbar-button-id",label:"New Mitosheet",title:"Create a blank Mitosheet below the active code cell",icon:"fa-regular fa-table",callback:()=>{(()=>{var e,o,t,n,i,l,r,d,u,a;null===(o=null===(e=window.Jupyter)||void 0===e?void 0:e.notebook)||void 0===o||o.insert_cell_below(),null===(n=null===(t=window.Jupyter)||void 0===t?void 0:t.notebook)||void 0===n||n.select_next();const s=null===(l=null===(i=window.Jupyter)||void 0===i?void 0:i.notebook)||void 0===l?void 0:l.get_cell(null===(d=null===(r=window.Jupyter)||void 0===r?void 0:r.notebook)||void 0===d?void 0:d.get_anchor_index());var v;void 0!==(v=s)&&""===function(e){return(null==e?void 0:e.get_text())||""}(v).trim()&&(function(e,o){null!=e&&e.set_text(o)}(s,"import mitosheet\nmitosheet.sheet()"),null===(a=null===(u=window.Jupyter)||void 0===u?void 0:u.notebook)||void 0===a||a.execute_cell_and_insert_below())})()}}]),e.exports={load_ipython_extension:function(){}}}},o={};function t(n){var i=o[n];if(void 0!==i)return i.exports;var l=o[n]={id:n,loaded:!1,exports:{}};return e[n](l,l.exports,t),l.loaded=!0,l.exports}return t.hmd=e=>((e=Object.create(e)).children||(e.children=[]),Object.defineProperty(e,"exports",{enumerable:!0,set:()=>{throw new Error("ES Modules may not assign module.exports or exports.*, Use ESM export syntax, instead: "+e.id)}}),e),t.r=e=>{"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},t(632)})()));